import jax
import jax.numpy as jnp
import jax.numpy as jnp
import jax.random as random
from flax import struct

import sys
sys.path.append('../../')

from structs import *
from functions import *

class Animal:
    pass
class Grass:
    pass



def interaction(wolves:Animal, sheeps:Animal, grasses:Grass):
    
    # wolves eat sheep:
    def one_wolf_all_sheep(wolf, sheeps): # vmap across all wolves

        def one_wolf_one_sheep(wolf, sheep): # vmap across all sheep
            wolf_X_pos = wolf.state.content['X_pos']
            wolf_Y_pos = wolf.state.content['Y_pos']
            sheep_X_pos = sheep.state.content['X_pos']
            sheep_Y_pos = sheep.state.content['Y_pos']
            condition = jnp.logical_and(wolf_X_pos[0] == sheep_X_pos[0], wolf_Y_pos[0] == sheep_Y_pos[0])
            
            wolf_energy_in = jax.lax.cond(condition, lambda _: jnp.array([1.0]), lambda _: jnp.array([0.0]), None)
            return wolf_energy_in
        
        one_wolf_energy_from_all_sheeps = jax.vmap(one_wolf_one_sheep, in_axes=(None, 0))(wolf, sheeps)
        return one_wolf_energy_from_all_sheeps
    
    wolves_sheeps_matrix = jax.vmap(one_wolf_all_sheep, in_axes=(0, None))(wolves, sheeps)
    '''
    for wolves this matrix is summed across all columns to get the total energy gained by each wolf
    for sheeps we take the max of the matrix across all rows to get if the sheep is eaten or not
    '''
    wolves_energy_in = jnp.sum(wolves_sheeps_matrix, axis=1, dtype=jnp.int32)
    sheeps_eaten = jnp.max(wolves_sheeps_matrix, axis=0)

    # sheeps eat grass:
    def one_sheep_all_grass(sheep, grasses): # vmap across all sheeps

        def one_sheep_one_grass(sheep, grass): # vmap across all grass
            sheep_X_pos = sheep.state.content['X_pos']
            sheep_Y_pos = sheep.state.content['Y_pos']
            grass_X_pos = grass.state.content['X_pos']
            grass_Y_pos = grass.state.content['Y_pos']
            grass_fully_grown = grass.state.content['fully_grown'][0] # fully grown = jnp.array([True]) or jnp.array([False])
            condition = jnp.logical_and(sheep_X_pos[0] == grass_X_pos[0], sheep_Y_pos[0] == grass_Y_pos[0])
            condition = jnp.logical_and(condition, grass_fully_grown)

            sheep_energy_in = jax.lax.cond(condition, lambda _: jnp.array([1.0]), lambda _: jnp.array([0.0]), None)
            return sheep_energy_in
        one_sheep_energy_from_all_grass = jax.vmap(one_sheep_one_grass, in_axes=(None, 0))(sheep, grasses)
        return one_sheep_energy_from_all_grass
    
    sheeps_grass_matrix = jax.vmap(one_sheep_all_grass, in_axes=(0, None))(sheeps, grasses)
    '''
    for sheeps this matrix is summed across all columns to get the total energy gained by each sheep
    for grass we take the max of the matrix across all rows to get if the grass is eaten or not
    '''
    sheeps_energy_in = jnp.sum(sheeps_grass_matrix, axis=1, dtype=jnp.int32)
    grasses_eaten = jnp.max(sheeps_grass_matrix, axis=0)

    return wolves_energy_in, sheeps_energy_in, sheeps_eaten, grasses_eaten
jit_interaction = jax.jit(interaction)


import jax
import jax.numpy as jnp
from flax import struct

import sys
sys.path.append('../../')

from structs import *
from functions import *

@struct.dataclass
class Order:
    pass

@struct.dataclass
class Trader:
    pass

def match_orders(buy_orders: Order, sell_orders: Order, traders: Trader):
    """
    Match buy and sell orders in the order book.
    This function matches buy and sell orders based on their prices and updates the traders' cash and shares accordingly.
    After sorting of buy and sell orders respectively, the matching is vectorized over the buy orders.
    Args:
        buy_orders (Order): Buy orders in the order book.
        sell_orders (Order): Sell orders in the order book.
        traders (Trader): Traders who are placing the orders.
    Returns:
        Tuple[Order, Order, Signal, Signal, jnp.ndarray, jnp.ndarray]:
            - buy_orders_sorted (Order): Sorted buy orders after matching.
            - sell_orders_sorted (Order): Sorted sell orders after matching.
            - buy_order_step_input (Signal): Signal containing the number of shares removed from each buy order.
            - sell_order_step_input (Signal): Signal containing the number of shares removed from each sell order.
            - traders_cash_change (jnp.ndarray): Change in cash for each trader.
            - traders_shares_change (jnp.ndarray): Change in shares for each trader.
    """
    # step 1 sort sell orders according to increasing price and buy orders according to decreasing price, first element is the best price
    buy_orders_sorted, b_indx = jit_sort_agents(quantity=-1*buy_orders.params.content["price"], agents=buy_orders)
    sell_orders_sorted, s_indx = jit_sort_agents(quantity=sell_orders.params.content["price"], agents=sell_orders)

    #step 2 compute the cumulative number of shares
    buy_cumulative_shares = jnp.cumsum(buy_orders_sorted.state.content["num_shares"]).reshape(-1)
    sell_cumulative_shares = jnp.cumsum(sell_orders_sorted.state.content["num_shares"]).reshape(-1)

    # step 3 calculate the transaction for each buy order
    def for_each_buy_order(buy_order, buy_cumulative_share, sell_orders, sell_cumulative_shares):
        
        #step 3.1 get max transaction price for the buy order and anything above that price is not reachable by the buy order
        transaction_mask = jnp.where(buy_order.params.content["price"] < sell_orders.params.content["price"],0,1).reshape(-1) #0-> not reachable, 1-> reachable
        
        # step 3.2 multiply the cumulative shares by the transaction mask, this will set the cumulative shares to 0 for all the unreachable orders
        sell_cumulative_shares = jnp.multiply(sell_cumulative_shares, transaction_mask) # this will set the cumulative shares to 0 for all the unreachable orders
            
        num_sell_order_shares = sell_orders.state.content["num_shares"].reshape(-1)
        num_sell_order_shares = jnp.multiply(num_sell_order_shares, transaction_mask)

        # step 3.3 now remove the share from the sell orders that will be matched by buy orders with more priority than the current buy order
        # for this we need to know how many shares have more priority than the current buy order, that = buy_cumulative_share - buy_order.state.content["num_shares"]
        num_higher_priority_shares = buy_cumulative_share - buy_order.state.content["num_shares"][0]
        
        new_sell_cumulative_shares = jnp.maximum(sell_cumulative_shares - num_higher_priority_shares, 0)
        new_num_sell_order_shares = jnp.minimum(num_sell_order_shares, new_sell_cumulative_shares)

        # step 3.5 now using the same method as above remove the shares of this buy order
        new_sell_cumulative_shares = jnp.maximum(new_sell_cumulative_shares - buy_order.state.content["num_shares"][0], 0)
        new_new_num_sell_order_shares = jnp.minimum(new_num_sell_order_shares, new_sell_cumulative_shares)

        share_change = new_num_sell_order_shares - new_new_num_sell_order_shares
        cash_change = jnp.multiply(share_change, sell_orders.params.content["price"].reshape(-1))

        return share_change, cash_change

    share_change, cash_change = jax.vmap(for_each_buy_order, in_axes=(0, 0, None, None))(buy_orders_sorted, buy_cumulative_shares, sell_orders_sorted, sell_cumulative_shares)

    # use change to caculate change in shares and cash for traders
    # change in buy and sell shares, each row is for a buy order and each column is for a sell order
    
    def for_each_trader(trader):
        buy_orders_mask = jnp.where(trader.id == buy_orders_sorted.params.content["trader_id"].reshape(-1), 1, 0)
        sell_orders_mask = jnp.where(trader.id == sell_orders_sorted.params.content["trader_id"].reshape(-1), 1, 0)

        # when buying, first mask then matrix -> mask x matrix multiplication
        cash_change_buy = jnp.sum(jnp.matmul(buy_orders_mask, cash_change))
        shares_change_buy = jnp.sum(jnp.matmul(buy_orders_mask, share_change))

        # when selling, first matrix then mask -> matrix x mask multiplication
        cash_change_sell = jnp.sum(jnp.matmul(cash_change, sell_orders_mask))
        shares_change_sell = jnp.sum(jnp.matmul(share_change, sell_orders_mask))

        # update the trader cash and shares change, when selling add cash and remove shares, when buying remove cash and add shares
        trader_cash_change = cash_change_sell - cash_change_buy
        trader_shares_change = shares_change_buy - shares_change_sell
        return trader_cash_change, trader_shares_change
    
    traders_cash_change, traders_shares_change = jax.vmap(for_each_trader)(traders)

    #traders_step_input = Signal(content={"cash_diff": traders_cash_change.reshape(-1, 1),"shares_diff": traders_shares_change.reshape(-1, 1)})
    
    buy_shares_removed = share_change.sum(axis=1).reshape(-1, 1)  # total shares removed for each buy order
    sell_shares_removed = share_change.sum(axis=0).reshape(-1, 1)  # total shares removed for each sell order

    buy_order_step_input = Signal(content={"num_shares_remove": buy_shares_removed})
    sell_order_step_input = Signal(content={"num_shares_remove": sell_shares_removed})

    
    return buy_orders_sorted, sell_orders_sorted, buy_order_step_input, sell_order_step_input, traders_cash_change, traders_shares_change


jit_match_orders = jax.jit(match_orders)

def select_traders(trading_flags):
    trading_flags = trading_flags.reshape(-1)
    selected_mask = jnp.where(trading_flags , 1, 0)
    sort_selected_indx = jnp.argsort(-1*selected_mask)
    num_selected = jnp.sum(selected_mask)
    return num_selected, sort_selected_indx
jit_select_traders = jax.jit(select_traders)

def get_price_list(price_list_arr, indx):
    '''
    Get the price list for the given indices
    '''
    return jnp.take(price_list_arr, indx).reshape(-1, 1)

def get_num_shares_list(num_shares_arr, indx):
    '''
    Get the number of shares list for the given indices
    '''
    return jnp.take(num_shares_arr, indx).reshape(-1, 1)

def get_empty_orders(orders:Agent, select_params:Params):
    return jnp.logical_not(orders.active_state)  # Selects the slots where orders are not active


def get_order_add_params_SCI(trader_set, buy_LOB, sell_LOB):
    num_buy_orders, buy_trader_id = jax.vmap(jit_select_traders, in_axes=(1))(trader_set.agents.state.content["buy_flag"])

    buy_price_list = jax.vmap(get_price_list, in_axes=(1,0))(trader_set.agents.state.content["buy_price"], buy_trader_id)
    buy_num_shares_list = jax.vmap(get_num_shares_list, in_axes=(1,0))(trader_set.agents.state.content["buy_num_shares"], buy_trader_id)

    num_sell_orders, sell_trader_id = jax.vmap(jit_select_traders, in_axes=(1))(trader_set.agents.state.content["sell_flag"])
    
    sell_price_list = jax.vmap(get_price_list, in_axes=(1,0))(trader_set.agents.state.content["sell_price"], sell_trader_id)
    sell_num_shares_list = jax.vmap(get_num_shares_list, in_axes=(1,0))(trader_set.agents.state.content["sell_num_shares"], sell_trader_id)

    num_empty_buy_orders, inactive_buy_order_index = jax.vmap(jit_select_agents, in_axes=(None, None, 0))(get_empty_orders, None, buy_LOB)
    num_empty_sell_orders, inactive_sell_order_index = jax.vmap(jit_select_agents, in_axes=(None, None, 0))(get_empty_orders, None, sell_LOB)

    num_buy_orders = jnp.minimum(num_buy_orders, num_empty_buy_orders)
    num_sell_orders = jnp.minimum(num_sell_orders, num_empty_sell_orders)

    buy_add_params = Params(content={"set_indx": inactive_buy_order_index,
                                     "price_list": buy_price_list,
                                     "num_shares_list": buy_num_shares_list, 
                                     "trader_id_list": buy_trader_id})

    sell_add_params = Params(content={"set_indx": inactive_sell_order_index,
                                      "price_list": sell_price_list,
                                      "num_shares_list": sell_num_shares_list,
                                      "trader_id_list": sell_trader_id})
    
    return buy_add_params, sell_add_params, num_buy_orders, num_sell_orders
jit_get_order_add_params = jax.jit(get_order_add_params_SCI)


import jax
import jax.numpy as jnp

from space.continuous.structs import *




def create_rays(agent_pos:tuple, ray_span = RAY_SPAN, ray_length = RAY_MAX_LENGTH):
    """
    Generate rays for the  agent given the agent position, the span of the rays, the length of the rays and the number of rays
    Args:
        - agent_pos: The position of the agent: (x, y, angle)
        - ray_span: The span of the rays i.e. how wide the agent can see times 0.5
        - ray_length: The length of the rays
        - RAY_RESOLUTION: The number of rays is global
    Returns:
        The rays generated by the foraer agent at the given position

    """
    x, y, angle = agent_pos

    ray_angles = jnp.linspace(angle - ray_span, angle + ray_span, RAY_RESOLUTION)
    cos_ray_angles = jnp.cos(ray_angles)
    sin_ray_angles = jnp.sin(ray_angles)
    ray_directions = jax.vmap(Point)(cos_ray_angles, sin_ray_angles)
    
    ray_origin = Point(x, y)
    
    rays = jax.vmap(Ray, in_axes=(None, 0, None))(ray_origin, ray_directions, ray_length)
    return rays

jit_create_rays = jax.jit(create_rays) #jitting is needed as rays are often generated at each instance of a loop


def create_space(x_min, x_max, y_min, y_max, torous, wall_array):
    """
    Create a space with the given parameters
    Args:
        - x_min: The minimum x value of the space
        - x_max: The maximum x value of the space
        - y_min: The minimum y value of the space
        - y_max: The maximum y value of the space
        - torous: Whether the space is torous or not
        - wall_array: The array of walls in the space in the form [[[wall1_begin(x,y)], [wall1_end(x,y)]], [[wall2_begin(x,y)], [wall2_end(x,y)], ...]]]
    Returns:
        A new space where wall is in the form of ([P1(x1,y1), P1(x2,y2), ..] [P2(x1,y1), P2(x2,y2), ..] ...)
    """
    x_wall_begins = jnp.array([wall_array[i][0][0] for i in range(len(wall_array))])
    y_wall_begins = jnp.array([wall_array[i][0][1] for i in range(len(wall_array))])
    x_wall_ends = jnp.array([wall_array[i][1][0] for i in range(len(wall_array))])
    y_wall_ends = jnp.array([wall_array[i][1][1] for i in range(len(wall_array))])
    wall_points_begins = jax.vmap(Point)(x_wall_begins, y_wall_begins)
    wall_points_ends = jax.vmap(Point)(x_wall_ends, y_wall_ends)
    walls = jax.vmap(Line)(wall_points_begins, wall_points_ends)
    return Space(x_min = x_min, x_max = x_max, y_min = y_min, y_max = y_max, torous = torous, walls = walls)


def is_walls_ray_collision(walls: Line, old_position:Point, new_position:Point):
    """
    Code to check if the particle collides with any of the walls
    code adapted from https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
    Args:
        - walls: The walls in the space
        - old_position: The old position of the particle
        - new_position: The new position of the particle
    Returns:
        - True if the particle collides with any of the walls, False otherwise
    """

    def is_wall_ray(wall, old_position, new_position):

        def is_on_segment(p, q, r):
            cond = jnp.logical_and(q.x <= jnp.maximum(p.x, r.x), 
                                   jnp.logical_and(q.x >= jnp.minimum(p.x, r.x), 
                                    jnp.logical_and(q.y <= jnp.maximum(p.y, r.y), q.y >= jnp.minimum(p.y, r.y))))
            return cond
    
        def get_orientation(p, q, r):
            val = (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)
            is_zero = jnp.isclose(val, 0.0, atol=1e-6)
            return jax.lax.select(is_zero, 0, jax.lax.select(val > 0, 1, 2))
        
        def is_intersect(wall, old_position, new_position):
            p1 = wall.p1
            q1 = wall.p2
            p2 = old_position
            q2 = new_position
            o1 = get_orientation(p1, q1, p2)
            o2 = get_orientation(p1, q1, q2)
            o3 = get_orientation(p2, q2, p1)
            o4 = get_orientation(p2, q2, q1)
            
            c1 = jnp.logical_and(o1 != o2, o3 != o4)
            c2 = jnp.logical_and(o1 == 0, is_on_segment(p1, p2, q1))
            c3 = jnp.logical_and(o2 == 0, is_on_segment(p1, q2, q1))
            c4 = jnp.logical_and(o3 == 0, is_on_segment(p2, p1, q2))
            c5 = jnp.logical_and(o4 == 0, is_on_segment(p2, q1, q2))
            return jnp.logical_or(jnp.logical_or(c1, c2), jnp.logical_or(c3, jnp.logical_or(c4, c5)))
        
        return is_intersect(wall, old_position, new_position)
    
    return jnp.any(jax.vmap(is_wall_ray, in_axes=(0, None, None))(walls, old_position, new_position)) # any rertuns true if any wall intersects with the particle

jit_is_walls_ray_collision = jax.jit(is_walls_ray_collision)



def get_ray_wall_collision(ray:Ray, wall:Line):
    """
    Step:
    - Get the distance of the collision of the ray with the walls along the ray
    - Def get_orientation(p, q, r): gets the orientation of the points p, q, r
    - Def get_line_intercept(p1, q1, p2, q2): gets the distance of the intersection point of the lines p1q1 and p2q2
    - Def is_on_segment(p, q, r): checks if the point q is on the segment pr
    Args:
        - ray: Ray, The ray to check for collision
        - walls: Line, The walls in the space
    Returns:
        The distance of the collision of the ray with the walls along the ray
    """

    def get_orientation(p, q, r):
        val = (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)
        is_zero = jnp.isclose(val, 0.0, atol=1e-6)
        return jax.lax.select(is_zero, 0, jax.lax.select(val > 0, 1, 2))
    
    def get_line_intercept(p1, q1, p2, q2):
        # converting points to arrays for cross product
        # p1 -> q1 is the ray, p2 -> q2 is the wall
        # algorithm adapted from https://stackoverflow.com/questions/20677795/how-do-i-compute-the-intersection-point-of-two-lines
        p1 = jnp.array([p1.x, p1.y])
        q1 = jnp.array([q1.x, q1.y])
        p2 = jnp.array([p2.x, p2.y])
        q2 = jnp.array([q2.x, q2.y])

        d1 = jnp.cross(q1 - p1, p2 - p1) # determinant of the ray with the (wall begin - ray begin)
        d2 = jnp.cross(q1 - p1, q2 - p1) # determinant of the ray with the (wall end - ray begin)
        denom = jax.lax.cond(d1 == d2, lambda _: 1e-6, lambda _: d1 - d2, None) # to avoid division by zero
        r = d1 / denom
        intersection_pt = p2 + r*(q2 - p2) # intersection point on the wall
        distance = jnp.linalg.norm(p1 - intersection_pt) # distance between the intersection point and the ray begin

        return distance
    
    def is_on_segment(p, q, r):
        # check if the point q is on the segment pr
        cond = jnp.logical_and(q.x <= jnp.maximum(p.x, r.x), 
                               jnp.logical_and(q.x >= jnp.minimum(p.x, r.x), 
                               jnp.logical_and(q.y <= jnp.maximum(p.y, r.y), q.y >= jnp.minimum(p.y, r.y))))
        return cond
    
    p1 = ray.origin
    q1 = Point(ray.origin.x + ray.direction.x*ray.length, ray.origin.y + ray.direction.y*ray.length)
    
    p2 = wall.p1
    q2 = wall.p2

    o_p1_q1_p2 = get_orientation(p1, q1, p2)
    o_p1_q1_q2 = get_orientation(p1, q1, q2)
    o_p2_q2_p1 = get_orientation(p2, q2, p1)
    o_p2_q2_q1 = get_orientation(p2, q2, q1)

    # the usual case
    c1 = jnp.logical_and(o_p1_q1_p2 != o_p1_q1_q2, o_p2_q2_p1 != o_p2_q2_q1) # if the orientations are different
    d1 = jax.lax.cond(c1, lambda _: get_line_intercept(p1, q1, p2, q2), lambda _: ray.length, None) # get the distance of the intersection point from the ray begin

    # if wall begin is on the ray
    c2 = jnp.logical_and(o_p1_q1_p2 == 0, is_on_segment(p1, p2, q1))
    d2 = jax.lax.cond(c2, lambda _: jnp.linalg.norm(jnp.array([p1.x, p1.y]) - jnp.array([p2.x, p2.y])), lambda _: ray.length, None)

    # if wall end is on the ray
    c3 = jnp.logical_and(o_p1_q1_q2 == 0, is_on_segment(p1, q2, q1))
    d3 = jax.lax.cond(c3, lambda _: jnp.linalg.norm(jnp.array([p1.x, p1.y]) - jnp.array([q2.x, q2.y])), lambda _: ray.length, None)

    # if ray begin is on the wall
    c4 = jnp.logical_and(o_p2_q2_p1 == 0, is_on_segment(p2, p1, q2))
    d4 = jax.lax.cond(c4, lambda _: 0.0, lambda _: ray.length, None)

    # return the minimum distance
    return jnp.min(jnp.array([d1, d2, d3, d4]))

jit_get_ray_wall_collision = jax.jit(get_ray_wall_collision)



def get_ray_circle_collision(ray:Ray, circle:Circle):
    """
    ray casting algorithm to check for collision between a ray and a circle, adapted from https://www.youtube.com/watch?v=ebzlMOw79Yw&ab_channel=MagellanicMath
    Args:
        - ray: Ray, The ray to check for collision
        - circle: Circle, The circle to check for collision
    Returns:
        The distance of the collision of the ray with the circle along the ray
    """
    circle_center = jnp.reshape(jnp.array([circle.center.x, circle.center.y]), (2,))
    ray_origin = jnp.reshape(jnp.array([ray.origin.x, ray.origin.y]), (2,))
    ray_direction = jnp.reshape(jnp.array([ray.direction.x, ray.direction.y]), (2,))

    s = ray_origin - circle_center
    b = jnp.dot(s, ray_direction)
    c = jnp.dot(s, s) - circle.radius**2
    h = b**2 - c
    h = jax.lax.cond(h < 0, lambda _: -1.0, lambda _: jnp.sqrt(h), None)
    t = jax.lax.cond(h >= 0, lambda _: -b - h, lambda _: ray.length, None)
    t = jax.lax.cond(t < 0, lambda _: ray.length, lambda _: t, None)
    return jnp.minimum(t, ray.length)

jit_get_ray_circle_collision = jax.jit(get_ray_circle_collision)

if __name__ == "__main__":
    agent_pos = (0.0, 0.0, 0.0)
    rays = jit_create_rays(agent_pos)
    

    wall_array = jnp.array([[[1.0,-1.0],[1.0,1.0]],[[1.0,1.0],[-1.0,1.0]],[[-1.0,1.0],[-1.0,-1.0]]]) # 3 sided wall
    space = create_space(-2.0, 2.0, -2.0, 2.0, True, wall_array)


    old_point = Point(0.0, 0.0)
    new_point = Point(2.0, 1.0)
    print(jit_is_walls_ray_collision(space.walls, old_point, new_point))

    old_point = Point(0.0, 0.0)
    new_point = Point(0.0, -2.0)
    print(jit_is_walls_ray_collision(space.walls, old_point, new_point))

    '''
    a_wall = jax.tree_util.tree_map(lambda x: x[0], space.walls)
    print(" a_wall start", a_wall.p1)
    print(" a_wall end", a_wall.p2)
    ray_wall_collisions = jax.vmap(jit_get_ray_wall_collision, in_axes=(0, None))(rays, a_wall)
    print(ray_wall_collisions)
    '''
    wall = Line(p1 = Point(1.0, -1.0), p2 = Point(1.0, 1.0))
    ray = Ray(origin = Point(0.0, 0.0), direction = Point(jnp.cos(0.0), jnp.sin(0.0)), length = 40.0)
    print(jit_get_ray_wall_collision(ray, wall))

    circle = Circle(center = Point(0.0, 0.0), radius = 1.0)



import jax
import jax.numpy as jnp
from flax import struct

from abmax.structs import *



def create_agents(agent:Agent, params:Params, num_agents:jnp.int32, num_active_agents:jnp.int32, agent_type:jnp.int16, key:jax.random.PRNGKey)->Agent:
    """
    Create agents with the given parameters
    assumption: create_agent method of the agent class is defined in the agent class by the user and returns an object of the agent class

    Args:
        agent: Agent class
        params: create parameters of the agent
        num_agents: Number of agents to create
        num_active_agents: Number of active agents
        agent_type: Type of agent
        key: Random key

    Returns:
        agents: object of Agent class, vmapped over params, ids, active_states, subkeys
    
    """
    ids = jnp.arange(num_agents)

    key, *subkeys = jax.random.split(key, num_agents+1)
    subkeys = jnp.array(subkeys)
    
    active_states = jnp.ones(num_agents)
    active_states = jnp.where(ids<num_active_agents, 1, 0)
    
    return jax.vmap(agent.create_agent, in_axes=(None, 0, 0, 0, 0))(agent_type, params, ids, active_states, subkeys)


def create_sets(set:Set, set_params:Params, set_type:jnp.int32,
                agent:Agent, agent_params:Params, agent_type:jnp.int32,
                num_sets:jnp.int32, num_agents:jnp.int32, num_active_agents:jnp.int32, key:jax.random.PRNGKey)->Set:
    """
    Create agent sets with the given parameters
    this is a vmapped version of Set.
    to create just one set, consider creating the set directly in the simulation
    assumption: create_set method of the set class is defined in the set class by the user and returns an object of the set class

    Args:
        set: Set class
        set_params: create parameters of the set (unique for each set, same for all the agents in the set)
        set_type: Type of set
        agent: Agent class
        agent_params: create parameters of the agents (unique for each agent, vmapped twice, once for each set and once for each agent)
        agent_type: Type of agent
        num_set: Number of sets to create
        num_agents: Number of agents to create
        num_active_agents: Number of active agents (different for each set)
        key: Random key


    Returns:
        sets: object of Set class
    
    """
    ids = jnp.arange(num_sets)

    key, *agent_subkeys = jax.random.split(key, num_sets+1)
    agent_subkeys = jnp.array(agent_subkeys)# each set has a different key
    
    key, *set_subkeys = jax.random.split(key, num_sets+1)
    set_subkeys = jnp.array(set_subkeys)
    
    agents = jax.vmap(create_agents, in_axes=(None, 0, None, 0, None, 0))(agent, agent_params, num_agents, num_active_agents, agent_type, agent_subkeys)
    
    return jax.vmap(set.create_set, in_axes=(None, 0, 0, 0, 0, None, 0))(num_agents, num_active_agents, agents, set_params, ids, set_type, set_subkeys)



def step_agents(step_func:callable, step_params:Params, input:Signal, set:Set)->Set:
    """
    Step the agents in the set with the given parameters and input
    assumption: step_func is defined in the agent class by the user and returns an object of the agent class
    Args:
        step_func: function to step the agents,
        step_params: parameters to step the agents, this is same for all the agents in the set
        input: input signal
        set: Set of agents
    returns:
        new set of agents, where the agents are replaced by the new agents
    """
    new_agents = jax.vmap(jax.jit(step_func), in_axes=(0, 0, None))(set.agents, input, step_params)
    return set.replace(agents=new_agents)

jit_step_agents = jax.jit(step_agents, static_argnums=(0,))


def set_agents_rank_match(set_func:callable, set_params:Params, mask_params:Params, num_agents:jnp.int32, set:Set)->Set:
    """
    Set particular agents in the set based on the rank match algorithm
    Make sure that the len(change_mask) == len(changes) in set_params.content
    assumption: set_func is defined in the agent class by the user and returns an object of the agent class

    args:
        set_func: function to set agents
        set_params: parameters to set agents, SHOULD HAVE 'set_indx' key in the content which
        contains the ids of the agents to set
        mask_params: parameters to set agents, SHOULD HAVE 'select_mask' and 'change_mask' keys in the content
        num_agents: number of agents in the set, this is used to limit the number of agents that can be set
        set: Set of agents
    returns:
        new set of agents, where the agents are replaced by the new agents
        num_changes: number of changes made, this is the number of agents that were set
    """
    select_mask = mask_params.content['select_mask']
    change_mask = mask_params.content['change_mask']
    

    cumsum_select_mask = jnp.cumsum(select_mask, dtype=jnp.int32)
    unique_select_mask = jnp.multiply(cumsum_select_mask, select_mask)

    cumsum_change_mask = jnp.cumsum(change_mask, dtype=jnp.int32)
    unique_change_mask = jnp.multiply(cumsum_change_mask, change_mask)

    num_changes = jnp.minimum(cumsum_select_mask[-1], cumsum_change_mask[-1]) # this is the number of changes that can be made

    def check_num_agent():
        return jnp.where(unique_change_mask > num_agents, 0, unique_change_mask), jnp.minimum(num_changes, num_agents)
    unique_change_mask, num_changes = jax.lax.cond(num_agents>-1, lambda _: check_num_agent(), lambda _: (unique_change_mask, num_changes), None)

    def set_agent(agent, select_mask_el):
        
        def check_mask(set_param, change_mask_el):
            def if_match():
                return jax.jit(set_func)(agent, set_param), 1
            def if_not_match():
                return agent, 0
            cond = jnp.logical_and(change_mask_el==select_mask_el, change_mask_el!=0)
            return jax.lax.cond(cond, lambda _: if_match(), lambda _: if_not_match(), None)
        
        agent_candidates, update_mask = jax.vmap(check_mask)(set_params, unique_change_mask)
        is_updated = jnp.sum(update_mask, dtype=jnp.int32) > 0
        update_index = jnp.argmax(update_mask)
        return jax.lax.cond(is_updated, lambda _: jax.tree_util.tree_map(lambda x: x[update_index], agent_candidates), lambda _: agent, None)

    new_agents = jax.vmap(set_agent)(set.agents, unique_select_mask)
    num_active_agents = jnp.sum(new_agents.active_state, dtype=jnp.int32)
    return set.replace(agents=new_agents, num_active_agents=num_active_agents), num_changes
jit_set_agents_rank_match = jax.jit(set_agents_rank_match, static_argnums=(0,))


def set_agents_mask(set_func:callable, set_params:Params, mask_params:Params, num_agents:jnp.int32, set:Set)->Set:
    """
    this function updates the agents when the changes are not unique, for eg removing agents or stepping agents
    mask is used to select the agents that need to be set
    assumption: set_func is defined in the agent class by the user and returns an object of the agent class
    args:
        set_func: function to set agents
        set_params: parameters to set agents
        mask_params: parameters to set agents, SHOULD HAVE 'set_mask' key in the content, highlighting the agents that need to be set
        num_agents: number of agents in the set, this is used to limit the number of agents that can be set
        set: Set of agents
    returns:
        new set of agents, where the agents are replaced by the new agents
        num_active_agents: number of active agents in the set after setting the agents
    """
    
    set_mask = mask_params.content['set_mask']
    unique_set_mask = jnp.multiply(jnp.cumsum(set_mask, dtype=jnp.int32), set_mask)

    unique_set_mask = jax.lax.cond(num_agents>-1,
        lambda _: jnp.where(unique_set_mask > num_agents, 0, unique_set_mask), 
        lambda _: unique_set_mask, 
        None
    )
    
    def set_agent(agent, set_mask_el, set_param):
        def if_set():
            return jax.jit(set_func)(agent, set_param)
        def if_not_set():
            return agent
        return jax.lax.cond(set_mask_el>0, lambda _: if_set(), lambda _: if_not_set(), None)

    new_agents = jax.vmap(set_agent, in_axes=(0, 0, None))(set.agents, unique_set_mask, set_params)
    num_active_agents = jnp.sum(new_agents.active_state, dtype=jnp.int32)
    return set.replace(agents=new_agents, num_active_agents=num_active_agents)

jit_set_agents_mask = jax.jit(set_agents_mask, static_argnums=(0,))



def set_agents_sci(set_func:callable, set_params:Params, num_agents_set:jnp.int32, set:Set)->Set:
    """
    apply unique updates to agents following the sort-count-iterate algorithm
    assumption: set_func is defined in the agent class by the user and returns an object of the agent class
    args:
        set_func: function to set agents
        num_agents_set: number of agents to set
        set_params: parameters to set agents, SHOULD HAVE 'set_indx' key in the content which
        contains the ids of the agents to set
        set: Set of agents
    returns:
        new set of agents, where the agents are replaced by the new agents
    """
    def set_data(idx, agents):
        set_indx = set_params.content['set_indx']
        new_agent = jax.jit(set_func)(agents, idx, set_params)
        new_agents = jax.tree_util.tree_map(lambda x,y:x.at[set_indx[idx]].set(y), agents, new_agent)
        return new_agents

    new_agents = jax.lax.fori_loop(0, num_agents_set, set_data, set.agents)
    return set.replace(agents=new_agents)
jit_set_agents_sci = jax.jit(set_agents_sci, static_argnums=(0,))


def sort_agents(quantity:jnp.array, agents:Agent)->Agent:
    """
    Sort the agents based on the quantity
    assumption: quantity is a 1D array of the same length as the number of agents

    args:
        quantity: quantity to sort the agents( multiply by -1 if you want to sort in descending order)
        agents: agents
    returns:
        new agents which are sorted based on the quantity
        sorted_indx: indexes of the agents in the ascending order
            for example consider quantity = [3, 1, 2]
            then the sorted_indx = [1, 2, 0] as 1 < 2 < 3 
    """
    quantity = jnp.reshape(quantity, (-1,))
    sorted_indx = jnp.argsort(quantity)
    new_agents = jax.tree_util.tree_map(lambda x: jnp.take(x, sorted_indx, axis=0), agents)
    return new_agents, sorted_indx

jit_sort_agents = jax.jit(sort_agents)


def sort_sets(quantity:jnp.array, sets:Set)->Set:
    """
    Sort the agents in the set based on the quantity
    assumption: quantity is a 1D array of the same length as the number of agents

    args:
        quantity: quantity to sort the agents
        set: Set of agents
    returns:
        new set of agents which are sorted based on the quantity
        sorted_indx: indexes of the agents in the ascending order
            for example consider quantity = [3, 1, 2]
            then the sorted_indx = [1, 2, 0] as 1 < 2 < 3

    """
    quantity = jnp.reshape(quantity, (-1,))
    sorted_indx = jnp.argsort(quantity)
    new_sets = jax.tree_util.tree_map(lambda x: jnp.take(x, sorted_indx, axis=0), sets)
    return new_sets, sorted_indx

jit_sort_sets = jax.jit(sort_sets)


def select_agents(select_func:bool, select_params:Params, set:Set)->tuple:
    """
    Select agents based on the select function
    assumption: select_func is defined by the user and 
    it takes in agents and select_params and returns a boolean array of the same length as the number of agents
    args:
        select_func: function to select agents
        select_params: parameters to select agents
        set: Set of agents
    returns:
        a tuple of:
        selected_indx_len: number of agents selected
        sort_selected_indx: indexes of the agents in the sorted order (descending order of active_state)
        So these 2 values can be used in a for loop to iterate over the selected agents
    """
    selected_indx = jnp.where(jax.jit(select_func)(set.agents, select_params), 1.0, 0.0)
    selected_indx = jnp.reshape(selected_indx,(-1,))

    sort_selected_indx = jnp.argsort(-1*selected_indx) # logic: 1 is greater than 0, so everywhere there is 1 will bubble up    
    selected_indx_len = jnp.sum(selected_indx, dtype=jnp.int32) # logic: sum of 1s is the number of selected agents
    return selected_indx_len, sort_selected_indx

jit_select_agents = jax.jit(select_agents, static_argnums=(0,))


def select_sets(select_func:bool, select_params:Params, set:Set)->tuple:
    """
    Select sets based on the select function
    assumption: select_func is defined by the user and
    it takes in agents and select_params and returns a boolean array of the same length as the number of sets

    args:
        select_func: function to select sets
        select_params: parameters to select sets
        set: Set of agents
    returns:
        a tuple of:
        selected_indx_len: number of sets selected
        sort_selected_indx: indexes of the sets in the sorted order (descending order of active_state)
        So these 2 values can be used in a for loop to iterate over the selected sets
    """
    selected_indx = jnp.where(jax.jit(select_func)(set, select_params), 1.0, 0.0)
    selected_indx = jnp.reshape(selected_indx,(-1,))
    sort_selected_indx = jnp.argsort(-1*selected_indx)
    selected_indx_len = jnp.sum(selected_indx, dtype=jnp.int32)
    return selected_indx_len, sort_selected_indx

jit_select_sets = jax.jit(select_sets, static_argnums=(0,))

def add_agents(add_func:callable, add_params:Params, num_agents_add:jnp.int32, set:Set)->Set:
    """
    Add agents to the set
    assumption: add_func is defined in the agent class by the user and returns an object of the agent class

    args:
        add_func: function to add agents
        num_agents_add: number of agents to add
        add_params: parameters to add agents
        set: Set of agents
    returns:
        new set of agents, where the agents are replaced by the new agents
    """
    id_last_active = set.num_active_agents
    max_agents_add = set.num_agents - id_last_active
    num_agents_add = jnp.minimum(num_agents_add, max_agents_add)

    def add_data(idx, agents):
        new_agent = jax.jit(add_func)(agents, idx, add_params)
        new_agents = jax.tree_util.tree_map(lambda x,y:x.at[idx].set(y), agents, new_agent)
        return new_agents
    
    new_agents = jax.lax.fori_loop(id_last_active, id_last_active+num_agents_add, add_data, set.agents)
    return set.replace(agents=new_agents, num_active_agents = set.num_active_agents + num_agents_add)
jit_add_agents = jax.jit(add_agents, static_argnums=(0,))


import jax
import jax.numpy as jnp
import jax.numpy as jnp
import jax.random as random
from flax import struct

import sys
sys.path.append('../../')

from structs import *
from functions import *

@struct.dataclass
class Cell:
    pass

@struct.dataclass
class Car:
    pass

@struct.dataclass
class Road:
    pass

def XY_to_cell_id(X:jnp.array, Y:jnp.array, X_max:jnp.array, Y_max:jnp.array):

    X_cond = jnp.logical_and(X[0] < X_max[0], X[0] >= 0)
    Y_cond = jnp.logical_and(Y[0] < Y_max[0], Y[0] >= 0)

    cell_id = jax.lax.cond(jnp.logical_and(X_cond, Y_cond), 
                            lambda _: X + jnp.multiply(Y, X_max), 
                            lambda _: jnp.array([-1]),
                            None)
    return cell_id
jit_XY_to_cell_id = jax.jit(XY_to_cell_id)

def cell_id_to_XY(cell_id, X_max, Y_max):
    X_cond = jnp.logical_and(cell_id[0] < jnp.multiply(X_max, Y_max)[0], cell_id[0] >= 0)
    Y_cond = jnp.logical_and(cell_id[0] < jnp.multiply(X_max, Y_max)[0], cell_id[0] >= 0)
    XY = jax.lax.cond(jnp.logical_and(X_cond, Y_cond), 
                      lambda _: (jnp.mod(cell_id, X_max), jnp.floor_divide(cell_id, X_max)),
                      lambda _: (jnp.array([-1]), jnp.array([-1])),
                      None)
    return XY
jit_cell_id_to_XY = jax.jit(cell_id_to_XY)


def is_move(cars: Agent, cells: Agent):
    #check if the car is chosen by the cell
    def is_chosen(car: Agent, cells: Agent):
        requested_cell_id = car.state.content['request_cell_id']
        current_cell_id = car.state.content['current_cell_id']
        exit = cells.params.content['exit'][current_cell_id[0]][0]  # Check if the current cell is an exit cell
        chosen_car_id = cells.state.content['car_id'][requested_cell_id][0]
        

        chosen = jax.lax.cond(jnp.logical_and(requested_cell_id[0] >= 0, exit==0), 
                              lambda _: car.id == chosen_car_id[0], # Only to move if the cell chose the car id and the cell is not an exit cell
                              lambda _: False, # If no cell has been requested yet, car is not to move.
                              None)
        return chosen

    chosen_arr = jax.vmap(is_chosen, in_axes=(0, None))(cars, cells)
    return chosen_arr
jit_is_move = jax.jit(is_move)

def is_remove(cars: Agent, cells:Agent):
    # if the car is at the exit cell, and the light is green(1), then it can be removed.
    def for_a_car(car):
        current_cell_id = car.state.content['current_cell_id']
        cond = jnp.logical_and(cells.params.content['exit'][current_cell_id[0]][0] == 1, cells.state.content['light'][current_cell_id[0]][0] == 1)
        cond = jnp.logical_and(cond, car.active_state) # car is active
        remove_flag = jax.lax.cond(cond,
                                  lambda _: 1, # car can be removed
                                  lambda _: 0, # car cannot be removed
                              None)
        return remove_flag
    remove_flags = jax.vmap(for_a_car)(cars)
    return remove_flags
jit_is_remove = jax.jit(is_remove)


def select_empty_slots(cars:Agent, select_params:Params):
    return jnp.logical_not(cars.active_state)  # Selects the slots where cars are not active

def spawn_cars(road:Road, key: jax.random.PRNGKey):
    cells = road.cells.agents

    entry_cell_ids = jnp.array([0,1,2])
    key, *subkey = random.split(key,3)

    num_cars_add = jax.random.randint(subkey[0], (1,), minval=1, maxval=4)[0]  # Randomly choose how many cars to spawn
    

    shuffled_entry_cell_ids = jax.random.permutation(subkey[1], entry_cell_ids)
    num_cars_shuffled = jnp.take(cells.state.content['num_cars'], shuffled_entry_cell_ids)
    is_cell_empty = jnp.where(num_cars_shuffled == 0, 1, 0)

    sorted_car_indexes  = jnp.argsort(-1*is_cell_empty.reshape(-1))  # Sort the car indexes based on the number of cars in the entry cells
    sorted_shuffled_entry_cell_ids = jnp.take(shuffled_entry_cell_ids, sorted_car_indexes)  # Get the sorted entry cell ids 

    num_empty_cells = jnp.sum(is_cell_empty)  # Calculate the number of empty cells in the entry cells
    num_cars_add = jnp.minimum(num_cars_add, num_empty_cells)  # Ensure we don't add more cars than empty cells

    _, inactive_car_index = jit_select_agents(select_empty_slots, None, road.cars)

    #define cell set params using set_sci method
    cell_set_params = Params(content={'set_indx': sorted_shuffled_entry_cell_ids, 'car_id': inactive_car_index})

    #define car add params based on the set_sci method
    car_add_params = Params(content={'set_indx': inactive_car_index, 'current_cell_ids': sorted_shuffled_entry_cell_ids})

    return cell_set_params, car_add_params, num_cars_add, key

jit_spawn_cars = jax.jit(spawn_cars)


from flax import struct
import jax
import jax.numpy as jnp

@struct.dataclass
class Point:
    """
    A point in 2D space.
    """
    x: jnp.float32
    y: jnp.float32

@struct.dataclass
class Line:
    """
    A line segment defined by two points.
    """
    p1: Point
    p2: Point

@struct.dataclass
class Circle:
    """
    A circle defined by its center:Point and radius.
    """
    center: Point
    radius: jnp.float32

@struct.dataclass
class Ray:
    """
    A ray defined by its origin:Point, direction:Point and length.
    direction is a unit vector (cos, sin)
    """
    origin: Point
    direction: Point # cos, sin
    length: float

# creating rays spontaneously uses jnp.arange which cannot be jit compiled thus we need to define these constants globally
RAY_RESOLUTION = 11 # number of rays
RAY_SPAN = jnp.pi/6
RAY_MAX_LENGTH = 40.0

@struct.dataclass
class Space:
    """
    A space defined by its bounds and walls.
    torous: bool determines if the space is torous or not i.e. exiting one side of the space will enter from the other side
    walls: Line is a vmapped array of walls in the space, A wall is defined by two points
    """
    x_min: jnp.float32
    x_max: jnp.float32
    y_min: jnp.float32
    y_max: jnp.float32
    torous: bool
    walls: Line # vmapped


from flax import struct
import jax
import jax.numpy as jnp

@struct.dataclass
class Signal:
    content:dict

@struct.dataclass
class Params:
    content:dict

@struct.dataclass
class State:
    content:dict

@struct.dataclass
class Policy:
    state:State
    params:Params
    key: jax.random.PRNGKey


@struct.dataclass
class Agent:
    id:jnp.int32
    active_state:bool
    age:jnp.float32
    agent_type:jnp.int32
    
    params:Params
    state:State
    policy:Policy
    key: jax.random.PRNGKey

@struct.dataclass
class Set:
    
    num_agents:jnp.int32
    num_active_agents:jnp.int32
    agents:Agent

    id:jnp.int32
    set_type:jnp.int32
    
    params:Params
    state:State
    policy:Policy
    key: jax.random.PRNGKey
    



from abmax.structs import *
from abmax.functions import *
import jax
import jax.random as random
import jax.numpy as jnp
from flax import struct



@struct.dataclass
class Continous_circular_forager(Agent):

    @staticmethod
    def create_agent(type, params, id, active_state, key):
        """
        The forager agent is created with the following design choices:
        - The forager is created with a random position in the space
        - The forager is created with a random velocity in the space
        - THe energy is initialized to energy_begin
        - The radius is initialized to radius and remains constant
        - The forager is created with a random policy

        state:
            - X: The x coordinate of the forager
            - Y: The y coordinate of the forager
            - X_dot: The x velocity of the forager
            - Y_dot: The y velocity of the forager
            - ang: The angle of the forager
            - energy: The energy of the forager
            - radius: The radius of the forager
        params:
            - space: The space in which the forager is created
            - energy_max: The maximum energy the forager can have
            - energy_min: The minimum energy the forager can have
            - energy_begin: The initial energy of the forager
            - radius: The radius of the forager
        Args:
            type: The type of the agent
            params: The parameters of the agent
            id: The id of the agent
            active_state: The active state of the agent
            key: The key used for random number generation
        Returns:
            A new forager agent
        """
        policy = params.content['policy']
        agent_params_content = {'space': params.content['space'], 'energy_max': params.content['energy_max'], 'energy_min': params.content['energy_min'], 'energy_begin': params.content['energy_begin']}
        agent_params = Params(content = agent_params_content)

        key, subkey = random.split(key)
        def create_active_state():
            _key, *create_keys = random.split(subkey, 5)
            X_len = params.content['space'].x_max - params.content['space'].x_min
            Y_len = params.content['space'].y_max - params.content['space'].y_min

            X = random.uniform(create_keys[0], (1,), minval=params.content['space'].x_min + 0.01*X_len, maxval=params.content['space'].x_max - 0.01*X_len)
            Y = random.uniform(create_keys[1], (1,), minval=params.content['space'].y_min + 0.01*Y_len, maxval=params.content['space'].y_max - 0.01*Y_len)
            vel_begin = params.content['vel_begin']
            X_dot = random.uniform(create_keys[2], (1,), minval=-vel_begin, maxval=vel_begin)
            Y_dot = random.uniform(create_keys[3], (1,), minval=-vel_begin, maxval=vel_begin)
            ang = jnp.arctan2(Y_dot, X_dot)

            energy = jnp.array([params.content['energy_begin']])
            radius = jnp.array([params.content['radius']])

            state_content = {'X': X, 'Y': Y, 'X_dot': X_dot, 'Y_dot': Y_dot, 'ang':ang, 'energy': energy, 'radius': radius}
            return State(content = state_content)
        
        def create_inactive_state():
            state_content = {'X': jnp.array([0.0]), 'Y': jnp.array([0.0]), 'X_dot': jnp.array([0.0]), 'Y_dot': jnp.array([0.0]), 'ang': jnp.array([0.0]), 
                             'energy': jnp.array([0.0]), 'radius': jnp.array([0.0])}
            return State(content = state_content)
        
        forager_state = jax.lax.cond(active_state, lambda _: create_active_state(), lambda _: create_inactive_state(), None)
        return Continous_circular_forager(id = id, active_state = active_state, age = 0.0, agent_type = type, params = agent_params, state = forager_state, policy = policy, key = key)
    



@struct.dataclass
class Continous_circular_patch(Agent):
    
    @staticmethod
    def create_agent(type, params, id, active_state, key):
        """
        Create a new patch agent with the given type, parameters, id, active state and key.
        Patch agent:
        state:
            - X: The x coordinate of the patch
            - Y: The y coordinate of the patch
            - energy: The energy of the patch
            - radius: The radius of the patch
            - energy_offer: The energy offer of the patch
        params:
            - eat_constant: The constant that determines how much energy is offered to the eater, energy_offer = eat_constant*energy
            - growth_rate: The rate at which the patch grows
            - energy_max: The maximum energy the patch can have
            - energy_min: The minimum energy the patch can have
            - space: The space in which the patch is created
        Design choices:
            - The patch is created with a random position in the space
            - All patches begin with the same energy and radius
            - radius remains constant
            - constant growth rate
        Args:
            - type: The type of the agent
            - params: The parameters of the agent
            - id: The id of the agent
            - active_state: The active state of the agent
            - key: The key used for random number generation
        Returns:
            A new patch agent
        """
        key, subkey = random.split(key)
        def create_active_state():
            _key, *create_keys = random.split(subkey,3)
        
            patch_radius = params.content['radius']
            space = params.content['space']
            space_x_len_margin = 0.01*(space.x_max - space.x_min) + patch_radius
            space_y_len_margin = 0.01*(space.y_max - space.y_min) + patch_radius

            X = random.uniform(create_keys[0], (1,), minval=space.x_min + space_x_len_margin, maxval=space.x_max - space_x_len_margin)
            Y = random.uniform(create_keys[1], (1,), minval=space.y_min + space_y_len_margin, maxval=space.y_max - space_y_len_margin)
            # We don't wall the patch to be to close to the edge of the space
            # right now uniform random initialization

            energy = jnp.array([params.content['energy_begin']])
            # all the patches start with the same energy

            radius = jnp.array([params.content['radius']])
            # all the patches have the same radius, it remains constant

            energy_offer = params.content['eat_constant']*energy
            # A certain amount of energy is up for grabs

            state_content = {'X': X, 'Y': Y, 'energy': energy, 'radius': radius, 'energy_offer': energy_offer}
            return State(content = state_content)
        
        # there are no inactive patches in the environment
        def create_inactive_state():
            state_content = {'X': jnp.array([0.0]), 'Y': jnp.array([0.0]), 'energy': jnp.array([0.0]), 'radius': jnp.array([0.0]), 'energy_offer': jnp.array([0.0])}
            return State(content = state_content)
        
        patch_state = jax.lax.cond(active_state, lambda _: create_active_state(), lambda _: create_inactive_state(), None)
        space = params.content['space']
        patch_params_content = {'eat_constant': params.content['eat_constant'], 'growth_rate': params.content['growth_rate'],
                                'energy_max': params.content['energy_max'], 'energy_min': params.content['energy_min'], 'x_max':space.x_max,
                                'y_max':space.y_max}
        # these params are apparently constants and not 1x1 arrays
        patch_params = Params(content = patch_params_content)

        return Continous_circular_patch(id = id, active_state = active_state, age = 0.0, agent_type = type, params = patch_params, state = patch_state, policy = None, key = key)
    


import jax
import jax.numpy as jnp
import jax.random as random


import sys
#sys.path.append('/Users/siddarth.chaturvedi/Desktop/source/abmax_git/abmax/')
from space.continuous.structs import *
from contexts.foraging.structs import *

def get_foragers_patches_interactions(foragers:Continous_circular_forager, patches:Continous_circular_patch):
    """
    steps:
    Calculate the energy_in and energy_out of the foragers and patches.active_state
        - Calculate the distance between the foragers and the patches. (each row is a forager, each column is a patch)
        - Calculate the energy that the foragers get from the patches. (each row is a forager, each column is a patch)
        - Foragers can get energy from multiple patches, thus sum the energy that the foragers get from the patches
        - A patch can feed multiple foragers, but in doing so it is depleted only once, by energy_offer amount. Thus take max of the energy matrix
    Args:
        - foragers: The foragers to calculate the energy_in and energy_out
        - patches: The patches to calculate the energy_in and energy_out
    Returns:
        - energy_in: The energy that the foragers get from the patches
        - energy_out: The energy that the patches give to

    """

    def forager_patches_dist(forager, patches):
        X_ps = patches.state.content['X'] # an array of shape (num_patches,)
        Y_ps = patches.state.content['Y']

        X_f = forager.state.content['X'] # a 1x1 array
        Y_f = forager.state.content['Y']
        
        return jnp.linalg.norm(jnp.stack((X_ps - X_f, Y_ps - Y_f), axis=1), axis=1)
    
    foragers_patches_dist = jax.vmap(forager_patches_dist, in_axes=(0, None))(foragers, patches)
    
    def forager_patches_energy(forager_patches_dist):
        def forager_patch_energy(dist, energy_offer, rad):
            return jax.lax.cond(dist[0] < rad[0], lambda _: energy_offer[0], lambda _: 0.0, None)
        return jax.vmap(forager_patch_energy, in_axes=(0, 0, 0))(forager_patches_dist, patches.state.content['energy_offer'], patches.state.content['radius'])
    
    foragers_patches_energy = jax.vmap(forager_patches_energy)(foragers_patches_dist)
    
    energy_in = jnp.sum(foragers_patches_energy, axis=1)# sum and not max because if there is overlapping resources, the forager can get energy from both
    energy_out = jnp.sum(foragers_patches_energy, axis=0)# does not matter as this is used as a boolean thus sum is the same as max
    
    return energy_in, energy_out

jit_get_foragers_patches_interactions = jax.jit(get_foragers_patches_interactions)


def get_sensor_data(foragers:Continous_circular_forager, patches:Continous_circular_patch, walls:Line):
    """
    Get the sensor data for the foragers. For each forager, for each ray -> [energy, type, distance]
    energy: The energy of the entity that the ray intercepts, nothing->0.0 wall -> 0.0, forager -> energy, patch -> energy
    type: The type of the entity that the ray intercepts, nothing->0, wall -> 1, forager -> 2, patch -> 3
    distance: The distance of the entity that the ray intercepts, nothing->ray_length, wall -> distance, forager -> distance, patch -> distance
    Args:
        foragers: The foragers to get the sensor data
        patches: The patches to get the sensor data
        walls: The walls in the space
    Returns:
        The sensor data for the foragers
    """
    
    agent_xs = jnp.concatenate([jnp.reshape(foragers.state.content['X'], (-1)), jnp.reshape(patches.state.content['X'], -1)])
    agent_ys = jnp.concatenate([jnp.reshape(foragers.state.content['Y'], (-1)), jnp.reshape(patches.state.content['Y'], -1)])
    agent_rads = jnp.concatenate([jnp.reshape(foragers.state.content['radius'], (-1)), jnp.reshape(patches.state.content['radius'], -1)])
    agent_types = jnp.concatenate([jnp.reshape(foragers.agent_type, (-1)), jnp.reshape(patches.agent_type, -1)])
    agent_active_states = jnp.concatenate([jnp.reshape(foragers.active_state, (-1)), jnp.reshape(patches.active_state, -1)])

    entity_types = jnp.concatenate([agent_types, jnp.ones_like(walls.p1.x)]) # 0->nothing, 1->wall, 2->forager, 3->patch
    entity_energies = jnp.concatenate([jnp.reshape(foragers.state.content['energy'], (-1)), jnp.reshape(patches.state.content['energy'], -1), jnp.zeros_like(walls.p1.x)])
    entity_energies = jnp.maximum(entity_energies, 0.0) # agents cant see negative energy
    #entity_ids = jnp.concatenate([jnp.reshape(foragers.id, (-1)), jnp.reshape(patches.id, -1), jnp.zeros_like(walls.p1.x)])
    env_states = jnp.vstack([entity_types, entity_energies]) # shape = (2, num_entities)

    #to speed up the computation, we can concatenate the wall information to the agent information
    data_ax_wp1x = jnp.concatenate([agent_xs, walls.p1.x]) # contains x coordinates of the agents and the beginning of the walls
    data_ay_wp1y = jnp.concatenate([agent_ys, walls.p1.y]) # contains y coordinates of the agents and the beginning of the walls
    data_ar_wp2x = jnp.concatenate([agent_rads, walls.p2.x]) # contains radii of the agents and the end of the walls
    data_aa_wp2y = jnp.concatenate([agent_active_states, walls.p2.y]) # contains active states of the agents and the end of the walls

    data_tuple = (data_ax_wp1x, data_ay_wp1y, data_ar_wp2x, data_aa_wp2y)
    
    def get_forager_sensor(forager):

        ray_span = RAY_SPAN # define the global variable
        ray_max_length = RAY_MAX_LENGTH # define the global variable
        forager_pos = (forager.state.content['X'][0], forager.state.content['Y'][0], forager.state.content['ang'][0]) #shape = (3,1)
        rays = generate_rays(forager_pos, ray_span, ray_max_length)

        def get_ray_sensor(ray):
        
            def get_ray_entity_sensor(data_tuple_row, entity_type):
                
                def ray_agent_collision():
                    circle = Circle(center = Point(data_tuple_row[0], data_tuple_row[1]), radius = data_tuple_row[2])
                    return jit_get_ray_circle_collision(ray, circle)
                
                def ray_wall_collision():
                    wall = Line(p1 = Point(data_tuple_row[0], data_tuple_row[1]), p2 = Point(data_tuple_row[2], data_tuple_row[3]))
                    return jit_get_ray_wall_collision(ray, wall)
                
                ray_collision = jax.lax.cond(entity_type == 1, lambda _: ray_wall_collision(), lambda _: ray_agent_collision(), None)
                return ray_collision
            
            ray_intercept_dist = jax.vmap(get_ray_entity_sensor)(data_tuple, entity_types)
            
            min_intercept_dist = jnp.min(ray_intercept_dist)
            min_intercept_index = jnp.argmin(ray_intercept_dist)
            min_intercept_index = jax.lax.cond(min_intercept_dist < ray_max_length, lambda _: min_intercept_index, lambda _: -1, None) # if the ray does not intercept with anything, return -1

            def get_sensor_data(env_states_row, min_intercept_index):
                return jax.lax.cond(min_intercept_index < 0, lambda _:0.0, lambda _: env_states_row[min_intercept_index], None) # 0.0 is energy value and type as wall is type 1
            
            sensor_data = jax.vmap(get_sensor_data, in_axes=(0, None))(env_states, min_intercept_index)# this vmap iterates over env_states = [entity_types, entity_energies], axis = 0
            
            ray_sensor= jnp.concatenate((sensor_data, jnp.array([min_intercept_dist])))

            return ray_sensor

        forager_sensor = jax.vmap(get_ray_sensor)(rays)
        return jnp.reshape(forager_sensor, (-1))
    
    foragers_sensor = jax.vmap(get_forager_sensor)(foragers)
    

    return foragers_sensor

        
jit_get_sensor_data = jax.jit(get_sensor_data)

